#include <cstdio>
#include <iostream>
#include <sys/poll.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <map>

#include "argh.hh"
#include "remotegcc.hh"

static unsigned port = 6580;

static vector<string> cmdline;
static const char *a0 = NULL;

static void Daemon(void)
{
	char portbuf[10];
	const char *argv[] =
	{
		"plis", "-s", portbuf, "-io", a0,
		NULL
	};
	
	sprintf(portbuf, "%u", port);
	
	plis_main(5, argv);
}

static void Output(FILE *fp)
{
    unsigned char Buf[16384];
    for(;;)
    {
    	int count = fread(Buf, 1, sizeof Buf, fp);
    	if(count <= 0)break;
    	for(unsigned char *s=Buf; count; ++s,--count)
    	{
    		if(*s == '\0') { putchar('\\'); putchar('0'); continue; }
    		if(*s == '\n') { putchar('\\'); putchar('b'); continue; }
    		if(*s == '\\') { putchar('\\'); putchar('\\'); continue; }
    		putchar(*s);
    	}
    }
    printf("\n");
}

static int Connection(void)
{
	map<string, unsigned> tmpfn;
	map<unsigned, string> fntrans;
	string s1, s2;
	int pip1[2];
	int pip2[2];
	int pid;
	
	// fprintf(stderr, "Connection!\n");
	
	unsigned fileind=3;
	string usestdin = "";
	for(;;)
	{
		getline(cin, s1);
		for(unsigned a=0; (a = s1.find('\r')) != s1.npos; )s1.erase(a, 1);
		if(!cin.good())break;
		// fprintf(stderr, "Got: '%s'\n", s1.c_str());
		if(s1==";")
		{
			if(!fileind)
			{
				if(usestdin.size())
				{
					getline(cin, s1, '\0');
					// fprintf(stderr, "Got enough (stdin size=%u)\n", s1.size());
					
					FILE *fp = fopen(usestdin.c_str(), "wb");
					fwrite(s1.c_str(), s1.size(), 1, fp);
					fclose(fp);
				}
				break;
			}
			fileind=0;
			continue;
		}
		if(fileind)
		{
			char Buf[32];
			sprintf(Buf, "/tmp/rgcc.%d.", getpid());
			tmpfn[s1] = fileind;
			fntrans[fileind] = Buf + s1;
		}
		else
		{
			map<string, unsigned>::const_iterator i;
			i = tmpfn.find(s1);
			if(i != tmpfn.end())
				s1 = fntrans[i->second];
			if(s1 == "-")
			{
				char Buf[32];
				sprintf(Buf, "/tmp/rgcc.%d", getpid());
				usestdin = Buf;
				usestdin += (cmdline[0]=="gcc") ? ".i" : ".ii";
				s1 = usestdin;
			}				 
			cmdline.push_back(s1);
		}
	}
	
	pipe(pip1);
	pipe(pip2);
	
	pid = fork();
	if(!pid)
	{
    	unsigned a, b=cmdline.size();
    	char **tab = new (char *) [b+1];
    	for(a=0; a<b; ++a)
    	{
    		const char *s = cmdline[a].c_str();
    		tab[a] = strdup(s);
    	}
    	tab[b] = NULL;
    	
    	close(0);
    	dup2(pip1[1], 1);
    	dup2(pip2[1], 2);
    	close(pip1[0]);
    	close(pip2[0]);
    	
	   	// for(unsigned a=0; tab[a]; ++a)fprintf(stderr, "Connection(): tab[%u] = \"%s\"\n", a, tab[a]);
    
    	if(execvp(tab[0], tab))
    	{
    		perror(tab[0]);
    		exit(-1);
    		delete[] tab;
    	}
	}
	close(pip1[1]);
	close(pip2[1]);
	
	struct pollfd fds[2];
	
	fds[0].fd = pip1[0]; fds[0].events = POLLIN | POLLHUP;
	fds[1].fd = pip2[0]; fds[1].events = POLLIN | POLLHUP;
	s1="";
	s2="";
	while(poll(fds, 2, -1) > 0)
	{
		char c;
		if(fds[0].revents & POLLIN) { read(fds[0].fd, &c, 1); if(c=='\n'){printf("1:%s\n",s1.c_str());fflush(stdout);s1="";}else s1+=c;}
		if(fds[1].revents & POLLIN) { read(fds[1].fd, &c, 1); if(c=='\n'){printf("2:%s\n",s2.c_str());fflush(stdout);s2="";}else s2+=c;}
		if((fds[0].revents & POLLHUP) && (fds[1].revents & POLLHUP))break;
	}
	kill(pid, SIGKILL);
	int status = 0;
	wait(&status);
	// fprintf(stderr, "State was 0x%X\n", status);
	printf("0:%u\n", WEXITSTATUS(status));
	
	if(usestdin.c_str())
		remove(usestdin.c_str());
	
	map<unsigned, string>::const_iterator i;
	for(i=fntrans.begin(); i!=fntrans.end(); ++i)
	{
		FILE *fp = fopen(i->second.c_str(), "rb");
		if(fp)
		{
    		printf("%u:", i->first);
    		
    		Output(fp);
    		fclose(fp);
    		remove(i->second.c_str());
		}
	}
	
	fflush(stdout);
	
	// fprintf(stderr, "Connection closed!\n");
	
	close(1);
	return 0;
}

static vector<string> hosts;
static vector<string> destfiles;
static bool Compile = false;
class Handle : public arghandler
{
	bool Help;
	bool Dest;
	
    string opt_h(const string &s)
    {
    	Help = true;
    	return s;
    }
    string opt_s(const string &s)
    {
    	// fprintf(stderr, "Adding host: \"%s\"\n", s.c_str());
    	hosts.push_back(s);
    	return "";
    }
public:
	Handle(const char *defopts, int argc, const char *const *argv)
	: arghandler(defopts, argc, argv),
		Help(false), Dest(false)
	{
		add("-h", "--help",     "This text", &(Handle::opt_h));
		add("-s", "--host",     "Add host",  &(Handle::opt_s));
	}
	virtual void defarg(const string &s)
	{
		if(Dest)
		{
			destfiles.push_back(s);
			Dest = false;
		}
		else if(s == "-o")
			Dest = true;
		else if(s == "-c")
			Compile = true;
		
		if(s.size())
			cmdline.push_back(s);
	}
    virtual void parse()
    {
    	arghandler::parse();
		if(Help)
		{
            printf("Usage: %s [options]\n", a0.c_str());
            
            printf("\nOptions:\n");
            
            listoptions();
            
            exit(0);
		}
    }
};

static string select_host(void)
{
	static unsigned long randseed;
	randseed = randseed * 0x8088401 + 1;
	return hosts[randseed % hosts.size()];
}

static void SafeSend(int fd, const string &s)
{
	const char *q = s.c_str();
	unsigned b = s.size();
	while(b)
	{
		int c = write(fd, q, b);
		if(c <= 0)break;
		b -= c;
		q += c;
	}		
}

static int Run(int argc, const char *const *argv)
{
	Handle parameters("", argc, argv);
	parameters.parse();
	
	string host = select_host();

 	char cbuf[16384];
 	
	int pip0[2];
	int pip1[2];
	int pid=0;
	
	string precompiled;
	
	if(Compile)
	{
	    pipe(pip1);
	    pid = fork();
	}
	
    if(!pid)
    {
      	unsigned b=cmdline.size();
      	char **tab = new (char *) [b+1];
      	unsigned c=0;
       	for(unsigned a=0; a<b; ++a)
       	{
      		const char *s = cmdline[a].c_str();
      		if(Compile)
      		{
          		if(cmdline[a] == "-c")s = "-E";
          		if(cmdline[a] == "-o")
          		{
          			++a;
          			continue;
          		}
      		}
      		tab[c++] = strdup(s);
      	}
      	tab[c] = NULL;
      	
      	if(Compile)
      	{
	      	close(0);
    	  	dup2(pip1[1], 1);
      		close(pip1[0]);
      	}
       	
       	// for(unsigned a=0; tab[a]; ++a)fprintf(stderr, "Run(): tab[%u] = \"%s\"\n", a, tab[a]);
          	
       	if(execvp(tab[0], tab))
       	{
       		perror(tab[0]);
       		exit(-1);
       		delete[] tab;
       	}
    }

    close(pip1[1]);
      	
 	for(int i; ((i = read(pip1[0], cbuf, sizeof cbuf)) > 0); )
 	{
 		cbuf[i] = 0;
  		precompiled += cbuf;
  	}
      	
    kill(pid, SIGKILL);

    int status;
    wait(&status);
    close(pip1[0]);

    if(WEXITSTATUS(status))
    	return WEXITSTATUS(status);
	
	pipe(pip0);
	pipe(pip1);
	
	pid = fork();
	if(!pid)
	{
    	char portbuf[10];
    	const char *argv[] =
    	{
    		"pcon", host.c_str(), portbuf, "-i", "cat", "-o", "cat",
    		NULL
    	};
    	
    	sprintf(portbuf, "%u", port);
    	
    	dup2(pip0[0], 0);
    	dup2(pip1[1], 1);
    	close(pip0[1]);
    	close(pip1[0]);
    	
    	// for(unsigned a=0; argv[a]; ++a)fprintf(stderr, "Run(): argv[%u] = \"%s\"\n", a, argv[a]);
    	
		_exit(plis_main(7, argv));
	}
	close(pip0[0]);
	close(pip1[1]);
	
	string s;
	
	for(unsigned a=0; a<destfiles.size(); ++a)
	{
		s += destfiles[a];
		s += '\n';
	}
	s += ";\n";
	bool outflag = false;
	for(unsigned a=0; a<cmdline.size(); ++a)
	{
		// fprintf(stderr, "Run(): cmdline[%u] = \"%s\"\n", a, cmdline[a].c_str());
		if(cmdline[a] == "-o")outflag = true;
		else if(outflag)outflag = false;
		else if(cmdline[a][0] != '-' && a>0)
			continue;
		s += cmdline[a];
		s += '\n';
	}
	s += "-\n;\n";
	// fprintf(stderr, "Precompiled is %u bytes\n", precompiled.size());
	s += precompiled;
	
	SafeSend(pip0[1], s);
	write(pip0[1], "", 1);
	close(pip0[1]);
	
	s = "";
	int exitvalue = 254;
	
	for(int i; (i=read(pip1[0], cbuf, sizeof cbuf)) > 0; )
		for(int ai=0; ai<i; ++ai)
		{
			char c = cbuf[ai];
    		if(c=='\n')
    		{
    			string s2 = s.substr(0, 2);
    			if(s2 == "0:")
    				exitvalue = atoi(s.substr(2).c_str());
    			else if(s2 == "1:")
    				fprintf(stdout, "%s\n", s.substr(2).c_str());
    			else if(s2 == "2:")
    				fprintf(stderr, "%s\n", s.substr(2).c_str());
    			else
    			{
    				const char *q = s.c_str();
    				int n=0;
    				while(*q && *q != ':')n = n*10 + *q++-'0';
    				if(*q == ':')++q;
    				if(n < 3)
    				{
    					fprintf(stderr, "remotegcc fatal error %d\n", n);
    					exit(-1);
    				}
    				FILE *fp = fopen(destfiles[n-3].c_str(), "wb");
    				if(fp)
    				{
    					while(*q)
    					{
    						if(*q == '\\')
    						{
    							switch(*++q)
    							{
    								case '0': putc('\0', fp); break;
    								case 'n': putc('\n', fp); break;
    								case '\\':putc('\\', fp); break;
    							}
    							++q; /* assume we got good data */
    						}
    						else
    							putc(*q++, fp);
    					}
    					fclose(fp);
    				}
    			}
    			s = "";
    		}
    		else
    			s += c;
    	}
	close(pip1[0]);
	kill(pid, SIGKILL);
	wait(NULL);
	
	return exitvalue;
}

int main(int argc, const char *const *argv)
{
	if(!argc)
		return Connection();

	a0 = argv[0];
	if(argc == 1)Daemon();
	return Run(argc, argv);
}

/*

	remotegcc -h oktober -c gcc -- -Wall -W -pedantic diu.c

*/
